"""A set of classes to build simple parent-children systems."""

import utils, IFakers

class FakeItem:
    """setters and getters for id and title"""

    __implements__=IFakers.IFakeItem

    def __init__(self, id, title=""):
        """an init"""
        self.setId(id)
        self.setTitle(title)

    def getTitle(self):
        """most basic title getter"""
        return self.title

    def setTitle(self, title):
        """sets title"""
        if self.getTitle()==title.strip(): return 0
        self.title=title.strip()
        return 1

    def title_or_id(self):
        """Utility that returns the title if it is not blank and the id
        otherwise."""
        return self.getTitle() or self.getId()

    def title_and_id(self):
        """Utility that returns the title if it is not blank and the id
        otherwise.  If the title is not blank, then the id is included
        in parens."""
        t=self.getTitle()
        return t and ("%s (%s)" % (t,self.getId())) or self.getId()

    def id_and_title(self):
        """Utility that returns the id, and title in parentheses if it
        is not blank."""
        t=self.getTitle()
        return t and ("%s (%s)" % (self.getId(), t)) or self.getId()

    def setId(self, id):
        """sets the id"""
        if self.getId()==id.strip(): return 0
        id=id.strip()
        msg=[]
        if not self.checkId(id, msg.append):
            raise ValueError, msg[-1]
        self.id=id.strip()
        return 1

    def getId(self):
        """basic getting the id"""
        return self.id

    _setId=setId
    id=getId

    def checkId(self, id, mfunc=None):
        """evaluates whether a given string is an acceptable id name"""
        if not id or type(id) != type(""):
            if mfunc: mfunc("cannot use an empty id")
            return 0
        elif id[0]=="_":
            if mfunc: mfunc("the id '%s' cannot be used because ids cannot begin with an underline character" % id)
            return 0
        else:
            return 1

class FakeProduct(FakeItem):
    """the most basic class for an object that can be a member of a FakeFolder"""

    __implements__=IFakers.IFakeProduct

    def __init__(self, id, title=""):
        """an init"""
        FakeItem.__init__(self, id, title)

    def getMetaType(self):
        """gets the meta_type of an object"""
        return self.meta_type

    def setMetaType(self, metaT):
        """sets the meta_type"""
        self.meta_type=metaT

    meta_type="Fake Product"

class FakeFolder( FakeItem ):
    """a simple folder for fake products"""

    __implements__=IFakers.IFakeFolder

    def __init__(self, id, title=""):
        """an init"""
        FakeItem.__init__(self, id, title)
        self._items=()

    addable_classes=(FakeProduct,)

    def getAddableClasses(self):
        """gets the addable classes available to this folder"""
        return self.addable_classes

    def setAddableClasses(self, classes):
        """sets this folder's addable classes (useful for self-referencing
        classes)"""
        if type(classes) != type(()):
            classes=(classes,)
        self.addable_classes=classes
        return 1

    def getAddableClassMetaTypes(self):
        """gets the metatypes of the available classes in this folder"""
        return [x.meta_type for x in self.getAddableClasses()]

    def getItems(self, metatypes=None, interfaces=None):
        """returns tuple of items"""
        if not metatypes and not interfaces:
            return self._items
        if type(metatypes) != type(()):
            metatypes=(metatypes,)
        if type(interfaces) != type(()):
            interfaces=(interfaces,)
        ret=[]
        for x in self._items:
            if metatypes and x.getMetaType() in metatypes:
                ret.append(x)
            elif interfaces:
                for ntrfc in interfaces:
                    if ntrfc.isImplementedBy(x):
                        ret.append(x)
                        break
        return ret

    def setItems(self, items):
        """completely resets the items list--only for private use"""
        tp=tuple(items)
        if self._items==tp: return 0
        self._items=tp
        return 1

    def getKeyValTuples(self, metatypes=None, interfaces=None):
        """returns a list of tuples, in which each tuple is the id and actual
        object of an item in the folder"""
        return [(x.getId(), x) for x in self.getItems(metatypes, interfaces)]

    def getItemById(self, id, marker=None):
        """pass in a string id, get the associated object or None if not
        found"""
        for item in self.getItems():
            if item.getId()==id: return item
        return marker

    def getItemByIdPath(self, path):
        """given a list of ids, tries to walk down it, finding the
        child known by the first id in the current object, the
        grandchild known by the second id in the child, and so on.  An
        error is raised if the path does not resolve to an object."""
        obj=self
        for id in path:
            if obj is utils.marker or not IFakers.IFakeFolder.isImplementedBy(obj):
                raise LookupError, "Path %s not found in fake folder %s" % ("/".join(path), self.getId())
            obj=obj.getItemById(id, utils.marker)
        if obj is utils.marker:
            raise LookupError, "Path %s not found in ZFakeFolder %s" % ("/".join(path), self.getId())
        return obj

    def getItemsByIds(self, ids=[]):
        """returns objects of given ids"""
        return [x for x in self.getItems() if x.getId() in ids]

    def getItemIds(self, metatypes=None, interfaces=None):
        """returns all ids of objects, restricted to metatypes and/or
        interfaces if given"""
        return [x.getId() for x in self.getItems(metatypes, interfaces)]

    def delItemsByIds(self, ids):
        """the nuts-and-bolts version of deleting items by a list or tuple of
        ids"""
        items=self.getItems()
        l=len(items)
        self.setItems([x for x in items if x.getId() not in ids])
        # self._p_changed = 1
        return l-len(self.getItems())


    def addItem(self, id, title="", meta_type=None, mfunc=None):
        """the nuts-and-bolts version of adding a single item.  Always adds to
        the bottom of the list for now."""
        classObj=None
        if not meta_type:
            if len(self.addable_classes)==1:
                classObj=self.addable_classes[0]
            elif self.addable_classes:
                if mfunc: mfunc("No class name specified")
                return 0
            else:
                if mfunc: mfunc("No addable classes defined for this folder")
                return 0
        else:
            for classObj in self.addable_classes:
                if hasattr(classObj,"meta_type") and classObj.meta_type == meta_type:
                    break
            else:
                if mfunc: mfunc("A '%s' cannot be added to this folder" % meta_type)
                return 0
        res=self.checkChildId(id, mfunc)
        if res:
            items=list(self.getItems())
            items.append(classObj(id, title)) # maybe change this to no args, but then setId and setTitle?
            return self.setItems(items)
        return 0

    def renameItemIds(self, oldNewIdPairs, mfunc=None):
        """the nuts-and-bolts version of renaming an item.  Expects a
        sequence of sequence pairs: oldId, newId.  Example:
        (("oldID","newID"),("anotherOld","anotherNew"))"""
        renames={}
        for oldId, id in oldNewIdPairs:
            if oldId != id:
                if self.checkId(id, mfunc):
                    item=self.getItemById(oldId)
                    if item:
                        item._setId(id)
                        renames[id]=oldId
            elif mfunc: mfunc("no change requested to %s" % id)
        return renames

    def checkChildId(self, id, mfunc=None):
        """evaluates whether a given string is an acceptable id name; extra check from original to check for duplicate ids"""
        #if hasattr(aq_base(self), id): # 1) aq_base not possible here 2) check unnecessary
        #    if mfunc: mfunc("'%s' is a reserved id" % id)
        #    return 0
        if self.getItemById(id):
            if mfunc: mfunc("an item identified as '%s' is already in this folder" % id)
            return 0
        elif " " in id:
            if mfunc: mfunc("this folder disallows ids with spaces in them such as '%s'" % id)
            return 0
        else:
            return 1

class OrderedFakeFolder( FakeFolder ):
    """An elaboration on the fakefolder that allows the items to have their order in the folder manipulated"""

    __implements__=IFakers.IOrderedFakeFolder

    def __init__(self, id, title=""):
        """an init"""
        FakeFolder.__init__(self, id, title)

    def moveItemsUpByIds(self, ids):
        """the nuts-and-bolts version of moving the items up"""
        self.setItems(utils.moveItemsUp(self.getItems(),ids))
        return 1

    def moveItemsDownByIds(self, ids):
        """the nuts-and-bolts version of moving the items down"""
        x=list(self.getItems())
        x.reverse()
        x=utils.moveItemsUp(x,ids)
        x.reverse()
        self.setItems(x)
        return 1

    def moveItemsToTopByIds(self, ids):
        """the nuts-and-bolts version of moving the items to the top"""
        mem, out=utils.separateItems(self.getItems(),ids)
        self.setItems(mem+out)
        return len(mem)

    def moveItemsToBottomByIds(self, ids):
        """the nuts-and-bolts version of moving the items to the bottom"""
        mem, out=utils.separateItems(self.getItems(),ids)
        self.setItems(out+mem)
        return len(mem)
